@@ -193,6 +193,10 @@ aggregator_putmetric(
 	struct _aggr_bucket *bucket;
 	struct _aggr_bucket_entries *entries;
 
+	/* do not accept new values when shutting down, issue #200 */
+	if (__sync_bool_compare_and_swap(&keep_running, 0, 0))
+		return;
+
 	/* get value */
 	if ((v = strchr(firstspace + 1, ' ')) == NULL) {
 		/* metric includes \n */
@@ -202,7 +206,7 @@ aggregator_putmetric(
 		return;
 	}
 
-	s->received++;
+	__sync_add_and_fetch(&s->received, 1);
 
 	val = atof(firstspace + 1);
 	epoch = atoll(v + 1);
@@ -237,21 +241,11 @@ aggregator_putmetric(
 			if (invocation->hash == omhash && \
 					strcmp(o, invocation->metric) == 0)  /* match */ \
 				break;
-		pthread_rwlock_rdlock(&compute->invlock);
-		find_invocation(ometric);
-		/* switch to a write lock from here, since we've found what we
-		 * were looking for (or are going to create it) and modify it */
-		pthread_rwlock_unlock(&compute->invlock);
 		pthread_rwlock_wrlock(&compute->invlock);
-
-		if (invocation == NULL) {
-			/* we need to recheck there wasn't someone else who did the
-			 * same thing we want to do below */
-			find_invocation(ometric);
-		}
+		find_invocation(ometric);
 
 		if (invocation == NULL) {  /* no match, add */
-			int i;
+			long long int i;
 			time_t now;
 
 			if ((invocation = malloc(sizeof(*invocation))) == NULL) {
@@ -295,7 +289,8 @@ aggregator_putmetric(
 				continue;
 			}
 			for (i = 0; i < s->bucketcnt; i++) {
-				invocation->buckets[i].start = now + (i * s->interval);
+				invocation->buckets[i].start =
+					now + (i * (long long int)s->interval);
 				invocation->buckets[i].cnt = 0;
 				invocation->buckets[i].entries.size = 0;
 				invocation->buckets[i].entries.values = NULL;
@@ -310,7 +305,7 @@ aggregator_putmetric(
 		itime = epoch - invocation->buckets[0].start;
 		if (itime < 0) {
 			/* drop too old metric */
-			s->dropped++;
+			__sync_add_and_fetch(&s->dropped, 1);
 			pthread_rwlock_unlock(&compute->invlock);
 			continue;
 		}
@@ -322,7 +317,7 @@ aggregator_putmetric(
 						"future (%lld > %lld): %s from %s", epoch,
 						invocation->buckets[s->bucketcnt - 1].start,
 						ometric, metric);
-			s->dropped++;
+			__sync_add_and_fetch(&s->dropped, 1);
 			pthread_rwlock_unlock(&compute->invlock);
 			continue;
 		}
@@ -403,14 +398,15 @@ aggregator_expire(void *sub)
 			/* send metrics for buckets that are completely past the
 			 * expiry time, unless we are shutting down, then send
 			 * metrics for all buckets that have completed */
-			now = time(NULL) + (keep_running ? 0 : s->expire - s->interval);
+			now = time(NULL) + (__sync_bool_compare_and_swap(&keep_running, 1, 1) ? 0 : s->expire - s->interval);
 			for (c = s->computes; c != NULL; c = c->next) {
+				pthread_rwlock_wrlock(&c->invlock);
 				for (i = 0; i < (1 << AGGR_HT_POW_SIZE); i++) {
 					lastinv = NULL;
 					isempty = 0;
 					for (inv = c->invocations_ht[i]; inv != NULL; ) {
 						while (inv->buckets[0].start +
-								(keep_running ? inv->expire : s->expire) < now)
+								(__sync_bool_compare_and_swap(&keep_running, 1, 1) ? inv->expire : s->expire) < now)
 						{
 							/* yay, let's produce something cool */
 							b = &inv->buckets[0];
@@ -501,19 +497,18 @@ aggregator_expire(void *sub)
 									logerr("aggregator: failed to write to "
 											"pipe (fd=%d): %s\n",
 											s->fd, strerror(errno));
-									s->dropped++;
+									__sync_add_and_fetch(&s->dropped, 1);
 								} else if (ts < len) {
 									logerr("aggregator: uncomplete write on "
 											"pipe (fd=%d)\n", s->fd);
-									s->dropped++;
+									__sync_add_and_fetch(&s->dropped, 1);
 								} else {
-									s->sent++;
+									__sync_add_and_fetch(&s->sent, 1);
 								}
 							}
 
 							/* move the bucket to the end, to make room for
 							 * new ones */
-							pthread_rwlock_wrlock(&c->invlock);
 							b = &inv->buckets[0];
 							len = b->entries.size;
 							values = b->entries.values;
@@ -526,7 +521,6 @@ aggregator_expire(void *sub)
 								s->interval;
 							b->entries.size = len;
 							b->entries.values = values;
-							pthread_rwlock_unlock(&c->invlock);
 
 							work++;
 						}
@@ -541,7 +535,6 @@ aggregator_expire(void *sub)
 							}
 						}
 						if (isempty) {
-							pthread_rwlock_wrlock(&c->invlock);
 							/* free and unlink */
 							if (c->entries_needed)
 								for (j = 0; j < s->bucketcnt; j++)
@@ -558,18 +551,18 @@ aggregator_expire(void *sub)
 								free(inv);
 								inv = c->invocations_ht[i];
 							}
-							pthread_rwlock_unlock(&c->invlock);
 						} else {
 							lastinv = inv;
 							inv = inv->next;
 						}
 					}
 				}
+				pthread_rwlock_unlock(&c->invlock);
 			}
 		}
 
 		if (work == 0) {
-			if (!keep_running)
+			if (__sync_bool_compare_and_swap(&keep_running, 0, 0))
 				break;
 			/* nothing done, avoid spinlocking */
 			usleep(250 * 1000);  /* 250ms */
@@ -664,7 +657,7 @@ aggregator_start(aggregator *aggrs)
 void
 aggregator_stop(void)
 {
-	keep_running = 0;
+	__sync_bool_compare_and_swap(&keep_running, 1, 0);
 	pthread_join(aggregatorid, NULL);
 }
 
@@ -677,7 +670,7 @@ aggregator_get_received(aggregator *a)
 	size_t totreceived = 0;
 
 	for ( ; a != NULL; a = a->next)
-		totreceived += a->received;
+		totreceived += __sync_add_and_fetch(&a->received, 0);
 
 	return totreceived;
 }
@@ -703,7 +696,7 @@ aggregator_get_sent(aggregator *a)
 	size_t totsent = 0;
 
 	for ( ; a != NULL; a = a->next)
-		totsent += a->sent;
+		totsent += __sync_add_and_fetch(&a->sent, 0);
 
 	return totsent;
 }
@@ -731,7 +724,7 @@ aggregator_get_dropped(aggregator *a)
 	size_t totdropped = 0;
 
 	for ( ; a != NULL; a = a->next)
-		totdropped += a->dropped;
+		totdropped += __sync_add_and_fetch(&a->dropped, 0);
 
 	return totdropped;
 }
