@@ -84,7 +84,7 @@ fnv1a_hashpos(const char *key, const char *end)
  * Computes the bucket number for key in the range [0, bckcnt).  The
  * algorithm used is the jump consistent hash by Lamping and Veach.
  */
-static unsigned int
+static int
 jump_bucketpos(unsigned long long int key, int bckcnt)
 {
 	long long int b = -1, j = 0;
@@ -481,8 +481,6 @@ ch_free(ch_ring *ring)
 
 	for (; ring->entries != NULL; ring->entries = ring->entries->next) {
 		if (ring->entries->malloced) {
-			free(ring->entries->server);
-
 			if (deletes == NULL) {
 				w = deletes = ring->entries;
 			} else {
@@ -491,12 +489,13 @@ ch_free(ch_ring *ring)
 		}
 	}
 
-	assert(w != NULL);
-	w->next = NULL;
-	while (deletes != NULL) {
-		w = deletes->next;
-		free(deletes);
-		deletes = w;
+	if (w != NULL) {
+		w->next = NULL;
+		while (deletes != NULL) {
+			w = deletes->next;
+			free(deletes);
+			deletes = w;
+		}
 	}
 
 	if (ring->entrylist != NULL)
