@@ -70,20 +70,22 @@ struct _dispatcher {
 	size_t prevblackholes;
 	size_t prevticks;
 	size_t prevsleeps;
-	char keep_running:1;
+	char keep_running;  /* full byte for atomic access */
 	router *rtr;
 	router *pending_rtr;
-	char route_refresh_pending:1;
+	char route_refresh_pending;  /* full byte for atomic access */
 	char hold:1;
 	char *allowed_chars;
 };
 
-static connection *listeners[32];       /* hopefully enough */
+static connection **listeners = NULL;
 static connection *connections = NULL;
 static size_t connectionslen = 0;
+pthread_rwlock_t listenerslock = PTHREAD_RWLOCK_INITIALIZER;
 pthread_rwlock_t connectionslock = PTHREAD_RWLOCK_INITIALIZER;
 static size_t acceptedconnections = 0;
 static size_t closedconnections = 0;
+static unsigned int sockbufsize = 0;
 
 
 /**
@@ -109,6 +111,8 @@ dispatch_check_rlimit_and_warn(void)
 	}
 }
 
+#define MAX_LISTENERS 32  /* hopefully enough */
+
 /**
  * Adds an (initial) listener socket to the chain of connections.
  * Listener sockets are those which need to be accept()-ed on.
@@ -126,43 +130,63 @@ dispatch_addlistener(int sock)
 	newconn->sock = sock;
 	newconn->takenby = 0;
 	newconn->buflen = 0;
-	for (c = 0; c < sizeof(listeners) / sizeof(connection *); c++)
-		if (__sync_bool_compare_and_swap(&(listeners[c]), NULL, newconn))
+
+	pthread_rwlock_wrlock(&listenerslock);
+	if (listeners == NULL) {
+		listeners = malloc(sizeof(connection *) * MAX_LISTENERS);
+		for (c = 0; c < MAX_LISTENERS; c++)
+			listeners[c] = NULL;
+	}
+	if (listeners == NULL) {
+		pthread_rwlock_unlock(&listenerslock);
+		return 1;
+	}
+	for (c = 0; c < MAX_LISTENERS; c++) {
+		if (listeners[c] == NULL) {
+			listeners[c] = newconn;
 			break;
-	if (c == sizeof(listeners) / sizeof(connection *)) {
+		}
+	}
+	if (c == MAX_LISTENERS) {
 		free(newconn);
 		logerr("cannot add new listener: "
 				"no more free listener slots (max = %zu)\n",
-				sizeof(listeners) / sizeof(connection *));
+				MAX_LISTENERS);
+		pthread_rwlock_unlock(&listenerslock);
 		return 1;
 	}
+	pthread_rwlock_unlock(&listenerslock);
 
 	return 0;
 }
 
+/**
+ * Remove listener from the listeners list.  Each removal will incur a
+ * global lock.  Frequent usage of this function is not anticipated.
+ */
 void
 dispatch_removelistener(int sock)
 {
 	int c;
 	connection *conn;
 
+	pthread_rwlock_wrlock(&listenerslock);
 	/* find connection */
-	for (c = 0; c < sizeof(listeners) / sizeof(connection *); c++)
+	for (c = 0; c < MAX_LISTENERS; c++)
 		if (listeners[c] != NULL && listeners[c]->sock == sock)
 			break;
-	if (c == sizeof(listeners) / sizeof(connection *)) {
+	if (c == MAX_LISTENERS) {
 		/* not found?!? */
 		logerr("dispatch: cannot find listener!\n");
+		pthread_rwlock_unlock(&listenerslock);
 		return;
 	}
-	/* make this connection no longer visible */
+	/* cleanup */
 	conn = listeners[c];
-	listeners[c] = NULL;
-	/* if some other thread was looking at conn, make sure it
-	 * will have moved on before freeing this object */
-	usleep(10 * 1000);  /* 10ms */
 	close(conn->sock);
 	free(conn);
+	listeners[c] = NULL;
+	pthread_rwlock_unlock(&listenerslock);
 }
 
 #define CONNGROWSZ  1024
@@ -235,6 +259,11 @@ dispatch_addconnection(int sock)
 	}
 
 	(void) fcntl(sock, F_SETFL, O_NONBLOCK);
+	if (sockbufsize > 0) {
+		if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
+				&sockbufsize, sizeof(sockbufsize)) != 0)
+			;
+	}
 	connections[c].sock = sock;
 	connections[c].buflen = 0;
 	connections[c].needmore = 0;
@@ -243,8 +272,9 @@ dispatch_addconnection(int sock)
 	connections[c].destlen = 0;
 	gettimeofday(&connections[c].lastwork, NULL);
 	connections[c].hadwork = 1;  /* force first iteration before stalling */
-	connections[c].takenby = 0;  /* now dispatchers will pick this one up */
-	acceptedconnections++;
+	/* after this dispatchers will pick this connection up */
+	__sync_bool_compare_and_swap(&(connections[c].takenby), -2, 0);
+	__sync_add_and_fetch(&acceptedconnections, 1);
 
 	return c;
 }
@@ -351,14 +381,14 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 
 	/* first try to resume any work being blocked */
 	if (dispatch_process_dests(conn, self, start) == 0) {
-		conn->takenby = 0;
+		__sync_bool_compare_and_swap(&(conn->takenby), self->id, 0);
 		return 0;
 	}
 
 	/* don't poll (read) when the last time we ran nothing happened,
 	 * this is to avoid excessive CPU usage, issue #126 */
 	if (!conn->hadwork && timediff(conn->lastwork, start) < 100 * 1000) {
-		conn->takenby = 0;
+		__sync_bool_compare_and_swap(&(conn->takenby), self->id, 0);
 		return 0;
 	}
 	conn->hadwork = 0;
@@ -396,15 +426,16 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 					continue;
 				}
 
-				self->metrics++;
+				__sync_add_and_fetch(&(self->metrics), 1);
 				*q++ = '\n';
 				*q = '\0';  /* can do this because we substract one from buf */
 
 				/* perform routing of this metric */
-				self->blackholes += router_route(self->rtr,
+				__sync_add_and_fetch(&(self->blackholes),
+						router_route(self->rtr,
 						conn->dests, &conn->destlen, CONN_DESTS_SIZE,
 						conn->srcaddr,
-						conn->metric, firstspace);
+						conn->metric, firstspace));
 
 				/* restart building new one from the start */
 				q = conn->metric;
@@ -471,7 +502,7 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 			/* force close connection below */
 			len = 0;
 		} else {
-			conn->takenby = 0;
+			__sync_bool_compare_and_swap(&(conn->takenby), self->id, 0);
 			return 0;
 		}
 	}
@@ -485,22 +516,22 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 			/* reset buffer only (UDP) and move on */
 			conn->needmore = 1;
 			conn->buflen = 0;
-			conn->takenby = 0;
+			__sync_bool_compare_and_swap(&(conn->takenby), self->id, 0);
 
 			return 0;
 		} else {
-			closedconnections++;
+			__sync_add_and_fetch(&closedconnections, 1);
 			close(conn->sock);
 
 			/* flag this connection as no longer in use */
-			conn->takenby = -1;
+			__sync_bool_compare_and_swap(&(conn->takenby), self->id, -1);
 
 			return 0;
 		}
 	}
 
 	/* "release" this connection again */
-	conn->takenby = 0;
+	__sync_bool_compare_and_swap(&(conn->takenby), self->id, 0);
 
 	return 1;
 }
@@ -516,26 +547,22 @@ dispatch_runner(void *arg)
 	connection *conn;
 	int c;
 
-	self->metrics = 0;
-	self->blackholes = 0;
-	self->ticks = 0;
-	self->sleeps = 0;
-	self->prevmetrics = 0;
-	self->prevblackholes = 0;
-	self->prevticks = 0;
-	self->prevsleeps = 0;
-
 	if (self->type == LISTENER) {
-		struct pollfd ufds[sizeof(listeners) / sizeof(connection *)];
-		while (self->keep_running) {
-			for (c = 0; c < sizeof(listeners) / sizeof(connection *); c++) {
+		struct pollfd ufds[MAX_LISTENERS];
+		int fds;
+		while (__sync_bool_compare_and_swap(&(self->keep_running), 1, 1)) {
+			pthread_rwlock_rdlock(&listenerslock);
+			fds = 0;
+			for (c = 0; c < MAX_LISTENERS; c++) {
 				if (listeners[c] == NULL)
-					break;
-				ufds[c].fd = listeners[c]->sock;
-				ufds[c].events = POLLIN;
+					continue;
+				ufds[fds].fd = listeners[c]->sock;
+				ufds[fds].events = POLLIN;
+				fds++;
 			}
-			if (poll(ufds, c, 1000) > 0) {
-				for (--c; c >= 0; c--) {
+			pthread_rwlock_unlock(&listenerslock);
+			if (poll(ufds, fds, 1000) > 0) {
+				for (c = fds - 1; c >= 0; c--) {
 					if (ufds[c].revents & POLLIN) {
 						int client;
 						struct sockaddr addr;
@@ -561,13 +588,13 @@ dispatch_runner(void *arg)
 		int work;
 		struct timeval start, stop;
 
-		while (self->keep_running) {
+		while (__sync_bool_compare_and_swap(&(self->keep_running), 1, 1)) {
 			work = 0;
 
-			if (self->route_refresh_pending) {
+			if (__sync_bool_compare_and_swap(&(self->route_refresh_pending), 1, 1)) {
 				self->rtr = self->pending_rtr;
 				self->pending_rtr = NULL;
-				self->route_refresh_pending = 0;
+				__sync_bool_compare_and_swap(&(self->route_refresh_pending), 1, 0);
 				self->hold = 0;
 			}
 
@@ -576,24 +603,28 @@ dispatch_runner(void *arg)
 			for (c = 0; c < connectionslen; c++) {
 				conn = &(connections[c]);
 				/* atomically try to "claim" this connection */
-				if (!__sync_bool_compare_and_swap(&(conn->takenby), 0, self->id))
+				if (!__sync_bool_compare_and_swap(
+							&(conn->takenby), 0, self->id))
 					continue;
 				if (self->hold && !conn->isaggr) {
-					conn->takenby = 0;
+					__sync_bool_compare_and_swap(
+							&(conn->takenby), self->id, 0);
 					continue;
 				}
 				work += dispatch_connection(conn, self, start);
 			}
 			pthread_rwlock_unlock(&connectionslock);
 			gettimeofday(&stop, NULL);
-			self->ticks += timediff(start, stop);
+			__sync_add_and_fetch(&(self->ticks), timediff(start, stop));
 
 			/* nothing done, avoid spinlocking */
-			if (self->keep_running && work == 0) {
+			if (__sync_bool_compare_and_swap(&(self->keep_running), 1, 1) &&
+					work == 0)
+			{
 				gettimeofday(&start, NULL);
 				usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
 				gettimeofday(&stop, NULL);
-				self->sleeps += timediff(start, stop);
+				__sync_add_and_fetch(&(self->sleeps), timediff(start, stop));
 			}
 		}
 	} else {
@@ -622,6 +653,16 @@ dispatch_new(char id, enum conntype type, router *r, char *allowed_chars)
 	ret->route_refresh_pending = 0;
 	ret->hold = 0;
 	ret->allowed_chars = allowed_chars;
+
+	ret->metrics = 0;
+	ret->blackholes = 0;
+	ret->ticks = 0;
+	ret->sleeps = 0;
+	ret->prevmetrics = 0;
+	ret->prevblackholes = 0;
+	ret->prevticks = 0;
+	ret->prevsleeps = 0;
+
 	if (pthread_create(&ret->tid, NULL, dispatch_runner, ret) != 0) {
 		free(ret);
 		return NULL;
@@ -637,9 +678,10 @@ static char globalid = 0;
  * (and putting them on the queue for handling the connections).
  */
 dispatcher *
-dispatch_new_listener(void)
+dispatch_new_listener(unsigned int nsockbufsize)
 {
 	char id = globalid++;
+	sockbufsize = nsockbufsize;
 	return dispatch_new(id, LISTENER, NULL, NULL);
 }
 
@@ -660,7 +702,7 @@ dispatch_new_connection(router *r, char *allowed_chars)
 void
 dispatch_stop(dispatcher *d)
 {
-	d->keep_running = 0;
+	__sync_bool_compare_and_swap(&(d->keep_running), 1, 0);
 }
 
 /**
@@ -704,7 +746,7 @@ inline void
 dispatch_schedulereload(dispatcher *d, router *r)
 {
 	d->pending_rtr = r;
-	d->route_refresh_pending = 1;
+	__sync_bool_compare_and_swap(&(d->route_refresh_pending), 0, 1);
 }
 
 /**
@@ -714,7 +756,7 @@ dispatch_schedulereload(dispatcher *d, router *r)
 inline char
 dispatch_reloadcomplete(dispatcher *d)
 {
-	return d->route_refresh_pending == 0;
+	return __sync_bool_compare_and_swap(&(d->route_refresh_pending), 0, 0);
 }
 
 /**
@@ -723,7 +765,7 @@ dispatch_reloadcomplete(dispatcher *d)
 inline size_t
 dispatch_get_ticks(dispatcher *self)
 {
-	return self->ticks;
+	return __sync_add_and_fetch(&(self->ticks), 0);
 }
 
 /**
@@ -733,7 +775,7 @@ dispatch_get_ticks(dispatcher *self)
 inline size_t
 dispatch_get_ticks_sub(dispatcher *self)
 {
-	size_t d = self->ticks - self->prevticks;
+	size_t d = dispatch_get_ticks(self) - self->prevticks;
 	self->prevticks += d;
 	return d;
 }
@@ -745,7 +787,7 @@ dispatch_get_ticks_sub(dispatcher *self)
 inline size_t
 dispatch_get_sleeps(dispatcher *self)
 {
-	return self->sleeps;
+	return __sync_add_and_fetch(&(self->sleeps), 0);
 }
 
 /**
@@ -755,7 +797,7 @@ dispatch_get_sleeps(dispatcher *self)
 inline size_t
 dispatch_get_sleeps_sub(dispatcher *self)
 {
-	size_t d = self->sleeps - self->prevsleeps;
+	size_t d = dispatch_get_sleeps(self) - self->prevsleeps;
 	self->prevsleeps += d;
 	return d;
 }
@@ -766,7 +808,7 @@ dispatch_get_sleeps_sub(dispatcher *self)
 inline size_t
 dispatch_get_metrics(dispatcher *self)
 {
-	return self->metrics;
+	return __sync_add_and_fetch(&(self->metrics), 0);
 }
 
 /**
@@ -776,7 +818,7 @@ dispatch_get_metrics(dispatcher *self)
 inline size_t
 dispatch_get_metrics_sub(dispatcher *self)
 {
-	size_t d = self->metrics - self->prevmetrics;
+	size_t d = dispatch_get_metrics(self) - self->prevmetrics;
 	self->prevmetrics += d;
 	return d;
 }
@@ -788,7 +830,7 @@ dispatch_get_metrics_sub(dispatcher *self)
 inline size_t
 dispatch_get_blackholes(dispatcher *self)
 {
-	return self->blackholes;
+	return __sync_add_and_fetch(&(self->blackholes), 0);
 }
 
 /**
@@ -798,7 +840,7 @@ dispatch_get_blackholes(dispatcher *self)
 inline size_t
 dispatch_get_blackholes_sub(dispatcher *self)
 {
-	size_t d = self->blackholes - self->prevblackholes;
+	size_t d = dispatch_get_blackholes(self) - self->prevblackholes;
 	self->prevblackholes += d;
 	return d;
 }
@@ -806,17 +848,17 @@ dispatch_get_blackholes_sub(dispatcher *self)
 /**
  * Returns the number of accepted connections thusfar.
  */
-size_t
+inline size_t
 dispatch_get_accepted_connections(void)
 {
-	return acceptedconnections;
+	return __sync_add_and_fetch(&(acceptedconnections), 0);
 }
 
 /**
  * Returns the number of closed connections thusfar.
  */
-size_t
+inline size_t
 dispatch_get_closed_connections(void)
 {
-	return closedconnections;
+	return __sync_add_and_fetch(&(closedconnections), 0);
 }
