@@ -30,7 +30,7 @@ int dispatch_addlistener_udp(int sock);
 void dispatch_removelistener(int sock);
 int dispatch_addconnection(int sock);
 int dispatch_addconnection_aggr(int sock);
-dispatcher *dispatch_new_listener(void);
+dispatcher *dispatch_new_listener(unsigned int sockbufsize);
 dispatcher *dispatch_new_connection(router *r, char *allowed_chars);
 void dispatch_stop(dispatcher *d);
 void dispatch_shutdown(dispatcher *d);
