@@ -38,14 +38,17 @@
 #include "aggregator.h"
 #include "collector.h"
 
-int keep_running = 1;
+unsigned char keep_running = 1;
+int pending_signal = -1;
 char relay_hostname[256];
 unsigned char mode = 0;
 
 static char *config = NULL;
 static int batchsize = 2500;
 static int queuesize = 25000;
+static int maxstalls = 4;
 static unsigned short iotimeout = 600;
+static int sockbufsize = 0;
 static dispatcher **workers = NULL;
 static char workercnt = 0;
 static router *rtr = NULL;
@@ -68,7 +71,7 @@ relaylog(enum logdst dest, const char *fmt, ...)
 	char prefix[64];
 	size_t len;
 	time_t now;
-	struct tm *tm_now;
+	struct tm tm_now;
 	FILE *dst = NULL;
 	char console = 0;
 	int ret;
@@ -88,12 +91,12 @@ relaylog(enum logdst dest, const char *fmt, ...)
 	assert(dst != NULL);
 
 	/* briefly stall if we're swapping fds */
-	while (!relay_can_log)
+	while (!__sync_add_and_fetch(&relay_can_log, 0))
 		usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
 
 	time(&now);
-	tm_now = localtime(&now);
-	len = strftime(prefix, sizeof(prefix), "[%Y-%m-%d %H:%M:%S]", tm_now);
+	localtime_r(&now, &tm_now);
+	len = strftime(prefix, sizeof(prefix), "[%Y-%m-%d %H:%M:%S]", &tm_now);
 
 	if (!console)
 		(void)snprintf(prefix + len, sizeof(prefix) - len, " (%s)", dest == LOGOUT ? "MSG" : "ERR");
@@ -111,33 +114,16 @@ relaylog(enum logdst dest, const char *fmt, ...)
 }
 
 static void
-exit_handler(int sig)
+sig_handler(int sig)
 {
-	char *sign = "unknown signal";
-
-	switch (sig) {
-		case SIGTERM:
-			sign = "SIGTERM";
-			break;
-		case SIGINT:
-			sign = "SIGINT";
-			break;
-		case SIGQUIT:
-			sign = "SIGQUIT";
-			break;
-	}
-	if (keep_running) {
-		logout("caught %s\n", sign);
-	} else {
-		logerr("caught %s while already shutting down, "
-				"forcing exit!\n", sign);
-		exit(1);
-	}
-	keep_running = 0;
+	/* a thunder of signals will be ignored for as long as a pending
+	 * signal was set, I can live with that, feels like a bad practice
+	 * to bombard a process like that */
+	(void)__sync_bool_compare_and_swap(&pending_signal, -1, sig);
 }
 
 static void
-hup_handler(int sig)
+do_reload(void)
 {
 	router *newrtr;
 	aggregator *newaggrs;
@@ -145,7 +131,6 @@ hup_handler(int sig)
 	FILE *newfd;
 	size_t numaggregators;
 
-	logout("caught SIGHUP...\n");
 	if (relay_stderr != stderr) {
 		/* try to re-open the file first, so we can still try and say
 		 * something if that fails */
@@ -154,24 +139,37 @@ hup_handler(int sig)
 					relay_logfile, strerror(errno));
 		} else {
 			logout("closing logfile\n");
-			relay_can_log = 0;
+			__sync_and_and_fetch(&relay_can_log, 0);
+			/* lame race avoidance for relaylog() usage */
+			usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
 			fclose(relay_stderr);
 			relay_stdout = newfd;
 			relay_stderr = newfd;
-			relay_can_log = 1;
+			__sync_bool_compare_and_swap(&relay_can_log, 0, 1);
 			logout("reopening logfile\n");
 		}
 	}
 
 	logout("reloading config from '%s'\n", config);
-	if ((newrtr = router_readconfig(config,
-					queuesize, batchsize, iotimeout)) == NULL)
+	if ((newrtr = router_readconfig(NULL, config,
+					queuesize, batchsize, maxstalls,
+					iotimeout, sockbufsize)) == NULL)
 	{
 		logerr("failed to read configuration '%s', aborting reload\n", config);
 		return;
 	}
 	router_optimise(newrtr);
 
+	/* compare the configs first, if there is no diff, then
+	 * just refrain from reloading anything */
+	if (!router_printdiffs(rtr, newrtr, relay_stdout)) {
+		/* no changes, shortcut */
+		logout("no config changes found\n");
+		router_free(newrtr);
+		logout("SIGHUP handler complete\n");
+		return;
+	}
+
 	logout("reloading collector\n");
 	collector_schedulereload(newrtr);
 	while (!collector_reloadcomplete())
@@ -207,6 +205,26 @@ hup_handler(int sig)
 		}
 	}
 
+	/* Transplant the queues for the same servers, so we keep their
+	 * queues (potentially with data) around.  To do so, we need to make
+	 * sure the servers aren't being operated on. */
+	router_shutdown(rtr);
+
+	/* Because we know that ip:port:prot is unique in both sets, when we
+	 * find a match, we can just swap the queue.  This comes with a
+	 * slight peculiarity, that if a server becomes part of an any_of
+	 * cluster (or failover), its pending queue can now be processed by
+	 * other servers.  While this feels wrong, it can be extremely
+	 * useful in case existing data needs to be migrated to another
+	 * server, e.g. the config on purpose added a failover to offload a
+	 * known dead server before removing it.  This is fairly specific
+	 * but in reality this happens to be desirable every once so often. */
+	router_transplant_queues(newrtr, rtr);
+
+	/* Before we start the workers sending traffic, start up the
+	 * servers now queues are transplanted. */
+	router_start(newrtr);
+
 	logout("reloading workers\n");
 	for (id = 1; id < 1 + workercnt; id++)
 		dispatch_schedulereload(workers[id], newrtr); /* un-holds */
@@ -222,6 +240,40 @@ hup_handler(int sig)
 	logout("SIGHUP handler complete\n");
 }
 
+static void
+handle_signal(int sig)
+{
+	char *sign = "unknown signal";
+
+	switch (sig) {
+		case SIGTERM:
+			sign = "SIGTERM";
+			break;
+		case SIGINT:
+			sign = "SIGINT";
+			break;
+		case SIGQUIT:
+			sign = "SIGQUIT";
+			break;
+		case SIGHUP:
+			sign = "SIGHUP";
+			break;
+	}
+	if (keep_running) {
+		logout("caught %s\n", sign);
+	} else {
+		/* also fine when SIGHUP */
+		logerr("caught %s while already shutting down, "
+				"forcing exit!\n", sign);
+		exit(1);
+	}
+	if (sig == SIGHUP) {
+		do_reload();
+	} else {
+		keep_running = 0;
+	}
+}
+
 static int
 get_cores(void)
 {
@@ -246,9 +298,9 @@ do_version(void)
 }
 
 static void
-do_usage(int exitcode)
+do_usage(char *name, int exitcode)
 {
-	printf("Usage: relay [-vdst] -f <config> [-p <port>] [-w <workers>] [-b <size>] [-q <size>]\n");
+	printf("Usage: %s -f <config> [opts]\n", name);
 	printf("\n");
 	printf("Options:\n");
 	printf("  -v  print version and exit\n");
@@ -257,11 +309,12 @@ do_usage(int exitcode)
 	printf("  -i  listen on <interface> for connections, defaults to all\n");
 	printf("  -l  write output to <file>, defaults to stdout/stderr\n");
 	printf("  -w  use <workers> worker threads, defaults to %d\n", get_cores());
-	printf("  -b  server send batch size, defaults to 2500\n");
-	printf("  -q  server queue size, defaults to 25000\n");
+	printf("  -b  server send batch size, defaults to %d\n", batchsize);
+	printf("  -q  server queue size, defaults to %d\n", queuesize);
+	printf("  -L  server max stalls, defaults to %d\n", maxstalls);
 	printf("  -S  statistics sending interval in seconds, defaults to 60\n");
-	printf("  -B  connection listen backlog, defaults to 3\n");
-	printf("  -T  IO timeout in milliseconds for server connections, defaults to 600\n");
+	printf("  -B  connection listen backlog, defaults to 32\n");
+	printf("  -T  IO timeout in milliseconds for server connections, defaults to %d\n", iotimeout);
 	printf("  -m  send statistics like carbon-cache.py, e.g. not cumulative\n");
 	printf("  -c  characters to allow next to [A-Za-z0-9], defaults to -_:#\n");
 	printf("  -d  debug mode: currently writes statistics to log, prints hash\n"
@@ -285,7 +338,7 @@ main(int argc, char * const argv[])
 	int dgram_socklen = sizeof(dgram_sock) / sizeof(dgram_sock[0]);
 	char id;
 	unsigned short listenport = 2003;
-	unsigned int listenbacklog = 3;
+	unsigned int listenbacklog = 32;
 	int ch;
 	size_t numaggregators;
 	char *listeninterface = NULL;
@@ -298,7 +351,7 @@ main(int argc, char * const argv[])
 	if (gethostname(relay_hostname, sizeof(relay_hostname)) < 0)
 		snprintf(relay_hostname, sizeof(relay_hostname), "127.0.0.1");
 
-	while ((ch = getopt(argc, argv, ":hvdmstf:i:l:p:w:b:q:S:T:c:H:B:DP:")) != -1) {
+	while ((ch = getopt(argc, argv, ":hvdmstf:i:l:p:w:b:q:L:S:T:c:H:B:U:DP:")) != -1) {
 		switch (ch) {
 			case 'v':
 				do_version();
@@ -328,28 +381,36 @@ main(int argc, char * const argv[])
 				listenport = (unsigned short)atoi(optarg);
 				if (listenport == 0) {
 					fprintf(stderr, "error: port needs to be a number >0\n");
-					do_usage(1);
+					do_usage(argv[0], 1);
 				}
 				break;
 			case 'w':
 				workercnt = (char)atoi(optarg);
 				if (workercnt <= 0) {
 					fprintf(stderr, "error: workers needs to be a number >0\n");
-					do_usage(1);
+					do_usage(argv[0], 1);
 				}
 				break;
 			case 'b':
 				batchsize = atoi(optarg);
 				if (batchsize <= 0) {
 					fprintf(stderr, "error: batch size needs to be a number >0\n");
-					do_usage(1);
+					do_usage(argv[0], 1);
 				}
 				break;
 			case 'q':
 				queuesize = atoi(optarg);
 				if (queuesize <= 0) {
 					fprintf(stderr, "error: queue size needs to be a number >0\n");
-					do_usage(1);
+					do_usage(argv[0], 1);
+				}
+				break;
+			case 'L':
+				maxstalls = atoi(optarg);
+				if (maxstalls < 0 || maxstalls >= (1 << SERVER_STALL_BITS)) {
+					fprintf(stderr, "error: maxium stalls needs to be a number "
+							"between 0 and %d\n", (1 << SERVER_STALL_BITS) - 1);
+					do_usage(argv[0], 1);
 				}
 				break;
 			case 'S':
@@ -357,17 +418,19 @@ main(int argc, char * const argv[])
 				if (collector_interval <= 0) {
 					fprintf(stderr, "error: sending interval needs to be "
 							"a number >0\n");
-					do_usage(1);
+					do_usage(argv[0], 1);
 				}
 				break;
 			case 'T': {
 				int val = atoi(optarg);
 				if (val <= 0) {
-					fprintf(stderr, "error: server IO timeout needs to be a number >0\n");
-					do_usage(1);
+					fprintf(stderr, "error: server IO timeout needs "
+							"to be a number >0\n");
+					do_usage(argv[0], 1);
 				} else if (val >= 60000) {
-					fprintf(stderr, "error: server IO timeout needs to be less than one minute\n");
-					do_usage(1);
+					fprintf(stderr, "error: server IO timeout needs "
+							"to be less than one minute\n");
+					do_usage(argv[0], 1);
 				}
 				iotimeout = (unsigned short)val;
 			}	break;
@@ -381,11 +444,46 @@ main(int argc, char * const argv[])
 				int val = atoi(optarg);
 				if (val <= 0) {
 					fprintf(stderr, "error: backlog needs to be a number >0\n");
-					do_usage(1);
+					do_usage(argv[0], 1);
 				}
 				listenbacklog = (unsigned int)val;
 			}	break;
+			case 'U': {
+				int val = atoi(optarg);
+				socklen_t len = sizeof(sockbufsize);
+				int sock;
 
+				if (val <= 0) {
+					fprintf(stderr, "error: bufsize needs to be a number >0\n");
+					do_usage(argv[0], 1);
+				}
+				sockbufsize = (unsigned int)val;
+
+				/* test the size to see if the value is valid, since
+				 * this is OS dependant, it's the only way */
+				sock = socket(PF_INET, SOCK_STREAM, 0);
+				if (sock == -1 || setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
+						&sockbufsize, sizeof(sockbufsize)) != 0)
+				{
+					fprintf(stderr, "failed to set socket bufsize: %s\n",
+							strerror(errno));
+				}
+				if (sock != -1 && getsockopt(sock, SOL_SOCKET, SO_RCVBUF,
+							&sockbufsize, &len) != 0)
+				{
+					fprintf(stderr, "failed to get socket bufsize: %s\n",
+							strerror(errno));
+				}
+				if (len != sizeof(sockbufsize)) {
+					fprintf(stderr, "getsockopt returned unexpected size: %u, "
+							"expected %lu\n", len, sizeof(sockbufsize));
+					exit(1);
+				}
+				if (sockbufsize != (unsigned int)val)
+					fprintf(stdout, "warning: OS rejected socket bufsize\n");
+				if (sock != -1)
+					close(sock);
+			}	break;
 			case 'D':
 				mode |= MODE_DAEMON;
 				break;
@@ -394,16 +492,16 @@ main(int argc, char * const argv[])
 				break;
 			case '?':
 			case ':':
-				do_usage(1);
+				do_usage(argv[0], 1);
 				break;
 			case 'h':
 			default:
-				do_usage(0);
+				do_usage(argv[0], 0);
 				break;
 		}
 	}
 	if (optind == 1 || config == NULL)
-		do_usage(1);
+		do_usage(argv[0], 1);
 
 
 	/* seed randomiser for dispatcher and aggregator "splay" */
@@ -480,15 +578,17 @@ main(int argc, char * const argv[])
 			if (p < 0) {
 				snprintf(msg, sizeof(msg), "failed to fork daemon process: %s",
 						strerror(errno));
-				write(fds[1], msg, strlen(msg) + 1);
-				exit(1); /* not that it matters */
+				/* fool compiler */
+				if (write(fds[1], msg, strlen(msg) + 1) > -2)
+					exit(1); /* not that retcode matters */
 			} else if (p == 0) {
 				/* child, this is the final daemon process */
 				if (setsid() < 0) {
 					snprintf(msg, sizeof(msg), "failed to setsid(): %s",
 							strerror(errno));
-					write(fds[1], msg, strlen(msg) + 1);
-					exit(1); /* not that it matters */
+					/* fool compiler */
+					if (write(fds[1], msg, strlen(msg) + 1) > -2)
+						exit(1); /* not that retcode matters */
 				}
 				/* close descriptors, we won't need them ever since
 				 * we're detached from the controlling terminal */
@@ -498,7 +598,8 @@ main(int argc, char * const argv[])
 
 				/* we're fine, flag our grandparent (= the original
 				 * process being called) it can terminate */
-				write(fds[1], "OK", 3);
+				if (write(fds[1], "OK", 3) < -2)
+					*msg = '\0';  /* unreachable dummy to fool compiler */
 			} else {
 				/* parent, die */
 				exit(0);
@@ -519,10 +620,16 @@ main(int argc, char * const argv[])
 				/* child is happy, so can we */
 				exit(0);
 			} else {
+				if (len == 1024)
+					len--;
+				msg[len] = '\0';
 				fprintf(stderr, "%s\n", msg);
 				exit(1);
 			}
 		}
+
+		close(fds[0]);
+		close(fds[1]);
 	}
 
 	if (pidfile_handle) {
@@ -541,9 +648,12 @@ main(int argc, char * const argv[])
 	fprintf(relay_stdout, "    workers = %d\n", workercnt);
 	fprintf(relay_stdout, "    send batch size = %d\n", batchsize);
 	fprintf(relay_stdout, "    server queue size = %d\n", queuesize);
+	fprintf(relay_stdout, "    server max stalls = %d\n", maxstalls);
 	fprintf(relay_stdout, "    statistics submission interval = %ds\n",
 			collector_interval);
 	fprintf(relay_stdout, "    listen backlog = %u\n", listenbacklog);
+	if (sockbufsize > 0)
+		fprintf(relay_stdout, "    socket bufsize = %u\n", sockbufsize);
 	fprintf(relay_stdout, "    server connection IO timeout = %dms\n",
 			iotimeout);
 	if (allowed_chars != NULL)
@@ -558,8 +668,9 @@ main(int argc, char * const argv[])
 	fprintf(relay_stdout, "    routes configuration = %s\n", config);
 	fprintf(relay_stdout, "\n");
 
-	if ((rtr = router_readconfig(config,
-					queuesize, batchsize, iotimeout)) == NULL)
+	if ((rtr = router_readconfig(NULL, config,
+					queuesize, batchsize, maxstalls,
+					iotimeout, sockbufsize)) == NULL)
 	{
 		logerr("failed to read configuration '%s'\n", config);
 		return 1;
@@ -596,19 +707,19 @@ main(int argc, char * const argv[])
 		exit(0);
 	}
 
-	if (signal(SIGINT, exit_handler) == SIG_ERR) {
+	if (signal(SIGINT, sig_handler) == SIG_ERR) {
 		logerr("failed to create SIGINT handler: %s\n", strerror(errno));
 		return 1;
 	}
-	if (signal(SIGTERM, exit_handler) == SIG_ERR) {
+	if (signal(SIGTERM, sig_handler) == SIG_ERR) {
 		logerr("failed to create SIGTERM handler: %s\n", strerror(errno));
 		return 1;
 	}
-	if (signal(SIGQUIT, exit_handler) == SIG_ERR) {
+	if (signal(SIGQUIT, sig_handler) == SIG_ERR) {
 		logerr("failed to create SIGQUIT handler: %s\n", strerror(errno));
 		return 1;
 	}
-	if (signal(SIGHUP, hup_handler) == SIG_ERR) {
+	if (signal(SIGHUP, sig_handler) == SIG_ERR) {
 		logerr("failed to create SIGHUP handler: %s\n", strerror(errno));
 		return 1;
 	}
@@ -643,7 +754,7 @@ main(int argc, char * const argv[])
 			return -1;
 		}
 	}
-	if ((workers[0] = dispatch_new_listener()) == NULL)
+	if ((workers[0] = dispatch_new_listener(sockbufsize)) == NULL)
 		logerr("failed to add listener\n");
 
 	if (allowed_chars == NULL)
@@ -664,12 +775,19 @@ main(int argc, char * const argv[])
 
 	/* server used for delivering metrics produced inside the relay,
 	 * that is, the collector (statistics) */
-	if ((internal_submission = server_new("internal", listenport, CON_PIPE,
-					NULL, 3000, batchsize, iotimeout)) == NULL)
+	if ((internal_submission = server_new(
+					"internal", listenport, CON_PIPE, NULL, 3000,
+					batchsize, maxstalls, iotimeout, sockbufsize)) == NULL)
 	{
 		logerr("failed to create internal submission queue, shutting down\n");
 		keep_running = 0;
 	}
+	if (internal_submission != NULL && server_start(internal_submission) != 0) {
+		logerr("failed to start submission queue thread, shutting down\n");
+		server_free(internal_submission);
+		internal_submission = NULL;
+		keep_running = 0;
+	}
 
 	if (numaggregators > 0) {
 		logout("starting aggregator\n");
@@ -680,13 +798,26 @@ main(int argc, char * const argv[])
 	}
 
 	logout("starting statistics collector\n");
-	collector_start(&workers[1], rtr, internal_submission, smode == CUM);
+	if (internal_submission != NULL)
+		collector_start(&workers[1], rtr, internal_submission, smode == CUM);
 
-	logout("startup sequence complete\n");
+	logout("starting servers\n");
+	if (router_start(rtr) != 0) {
+		/* router_start logerrs itself */
+		keep_running = 0;
+	}
+
+	if (keep_running == 0)
+		logout("startup sequence complete\n");
 
 	/* workers do the work, just wait */
-	while (keep_running)
+	while (keep_running) {
 		sleep(1);
+		if (pending_signal != -1) {
+			handle_signal(pending_signal);
+			pending_signal = -1;
+		}
+	}
 
 	logout("shutting down...\n");
 	/* make sure we don't accept anything new anymore */
@@ -720,6 +851,7 @@ main(int argc, char * const argv[])
 	fflush(relay_stdout);
 	free(workers);
 
+	router_shutdown(rtr);
 	router_free(rtr);
 	logout("stopped servers\n");
 
