@@ -34,8 +34,11 @@ typedef struct _router router;
 
 #define RE_MAX_MATCHES     64
 
-router *router_readconfig(const char *path, size_t queuesize, size_t batchsize, unsigned short iotimeout);
+router *router_readconfig(router *orig, const char *path, size_t queuesize, size_t batchsize, int maxstalls, unsigned short iotimeout, unsigned int sockbufsize);
 void router_optimise(router *r);
+char router_printdiffs(router *old, router *new, FILE *out);
+void router_transplant_queues(router *new, router *old);
+char router_start(router *r);
 size_t router_rewrite_metric(char (*newmetric)[METRIC_BUFSIZ], char **newfirstspace, const char *metric, const char *firstspace, const char *replacement, const size_t nmatch, const regmatch_t *pmatch);
 void router_printconfig(router *r, FILE *f, char mode);
 char router_route(router *r, destination ret[], size_t *retcnt, size_t retsize, char *srcaddr, char *metric, char *firstspace);
@@ -43,6 +46,7 @@ void router_test(router *r, char *metric_path);
 server **router_getservers(router *r);
 aggregator *router_getaggregators(router *r);
 char *router_getcollectorstub(router *r);
+void router_shutdown(router *r);
 void router_free(router *r);
 
 #endif
